from typing import Callable

from src.models.node import (
  ExecFn,
  FnLib,
  HOExecFn,
  HOLib,
)

def rename_fn(name: str) -> Callable[[ExecFn], ExecFn]:
  def decorator(fn: ExecFn) -> ExecFn:
    fn.__name__ = name
    return fn
  
  return decorator

def rename_ho(name: str) -> Callable[[HOExecFn], HOExecFn]:
  def decorator(fn: HOExecFn) -> HOExecFn:
    fn.__name__ = name
    return fn
  
  return decorator

def register_exec_fn(
  lib: FnLib,
) -> Callable[[ExecFn], ExecFn]:
  def decorator(fn: ExecFn) -> ExecFn:
    lib[fn.__name__] = fn
    return fn
  
  return decorator

def register_ho_fn(
  lib: HOLib,
) -> Callable[[HOExecFn], HOExecFn]:
  def decorator(fn: HOExecFn) -> HOExecFn:
    lib[fn.__name__] = fn
    return fn
  
  return decorator

def register_fn(
  lib: list[ExecFn]
) -> Callable[[ExecFn], ExecFn]:
  def decorator(fn: ExecFn) -> ExecFn:
    lib.append(fn)
    return fn
  
  return decorator

def register_ho(
  lib: list[HOExecFn]
) -> Callable[[HOExecFn], HOExecFn]:
  def decorator(fn: HOExecFn) -> HOExecFn:
    lib.append(fn)
    return fn
  
  return decorator

def setup(
  fn_lib: FnLib,
  ho_lib: HOLib,
  registered_exec_fns: list[ExecFn],
  registered_ho_fns: list[HOExecFn],
) -> None:
  for fn in registered_exec_fns:
    register_exec_fn(fn_lib)(fn)

  for fn in registered_ho_fns:
    register_ho_fn(ho_lib)(fn)




from src.models.node_type import NodeType
from src.models.base_node import BaseNode
from src.models.node import Node, FnLib, ExecFn, HOLib, HOExecFn

from src.internals.registry import (
  register_fn,
  rename_fn,
  setup as setup_registry
)

fn_exports: list[ExecFn] = []
ho_exports: list[HOExecFn] = []
def setup(fn_lib: FnLib, ho_lib: HOLib) -> None:
  setup_registry(fn_lib, ho_lib, fn_exports, ho_exports)

@register_fn(fn_exports)
def add(
  args: list[BaseNode],
  _: Node | None = None,
  __: FnLib = {},
) -> BaseNode:
  result: float = 0.0
  for arg in args:
    result += arg.fetch_float()
  
  return BaseNode(
    type=NodeType.NUMERIC,
    float_value=result
  )


@register_fn(fn_exports)
def multiply(
  args: list[BaseNode],
  _: Node | None = None,
  __: FnLib = {},
) -> BaseNode:
  result: float = 1.0
  for arg in args:
    result *= arg.fetch_float()
  
  return BaseNode(
    type=NodeType.NUMERIC,
    float_value=result
  )

@register_fn(fn_exports)
def subtract(
  args: list[BaseNode],
  _: Node | None = None,
  __: FnLib = {},
) -> BaseNode:
  args.append(BaseNode(type=NodeType.NUMERIC, float_value=0.0))
  args.append(BaseNode(type=NodeType.NUMERIC, float_value=0.0))
  
  result: float = args[0].fetch_float() - args[1].fetch_float()
  return BaseNode(
    type=NodeType.NUMERIC,
    float_value=result
  )

@register_fn(fn_exports)
def divide(
  args: list[BaseNode],
  _: Node | None = None,
  __: FnLib = {},
) -> BaseNode:
  args.append(BaseNode(type=NodeType.NUMERIC, float_value=0.0))
  args.append(BaseNode(type=NodeType.NUMERIC, float_value=1.0))
  
  result: float = args[0].fetch_float() / args[1].fetch_float()
  return BaseNode(
    type=NodeType.NUMERIC,
    float_value=result
  )

@register_fn(fn_exports)
def equals(
  args: list[BaseNode],
  _: Node | None = None,
  __: FnLib = {},
) -> BaseNode:
  if len(args) < 2:
    return BaseNode(type=NodeType.BOOLEAN, bool_value=False)
  
  types = set(arg.type for arg in args)
  if len(types) > 1:
    return BaseNode(type=NodeType.BOOLEAN, bool_value=False)
  
  if NodeType.NUMERIC in types:
    result: bool = args[0].fetch_float() == args[1].fetch_float()
  elif NodeType.BOOLEAN in types:
    result: bool = args[0].fetch_bool() == args[1].fetch_bool()
  else:
    result: bool = args[0].fetch_str() == args[1].fetch_str()

  return BaseNode(
    type=NodeType.BOOLEAN,
    bool_value=result
  )

@register_fn(fn_exports)
def not_equals(
  args: list[BaseNode],
  _: Node | None = None,
  __: FnLib = {},
) -> BaseNode:
  return BaseNode(
    type=NodeType.BOOLEAN,
    bool_value=not equals(args, None, __).fetch_bool()
  )

@register_fn(fn_exports)
def less_than(
  args: list[BaseNode],
  _: Node | None = None,
  __: FnLib = {},
) -> BaseNode:
  if len(args) < 2:
    return BaseNode(type=NodeType.BOOLEAN, bool_value=False)
  
  result: bool = args[0].fetch_float() < args[1].fetch_float()
  return BaseNode(
    type=NodeType.BOOLEAN,
    bool_value=result
  )

@register_fn(fn_exports)
def greater_than(
  args: list[BaseNode],
  _: Node | None = None,
  __: FnLib = {},
) -> BaseNode:
  if len(args) < 2:
    return BaseNode(type=NodeType.BOOLEAN, bool_value=False)
  
  result: bool = args[0].fetch_float() > args[1].fetch_float()
  return BaseNode(
    type=NodeType.BOOLEAN,
    bool_value=result
  )

@register_fn(fn_exports)
def less_than_equal(
  args: list[BaseNode],
  _: Node | None = None,
  __: FnLib = {},
) -> BaseNode:
  if len(args) < 2:
    return BaseNode(type=NodeType.BOOLEAN, bool_value=False)
  
  result: bool = args[0].fetch_float() <= args[1].fetch_float()
  return BaseNode(
    type=NodeType.BOOLEAN,
    bool_value=result
  )

@register_fn(fn_exports)
def greater_than_equal(
  args: list[BaseNode],
  _: Node | None = None,
  __: FnLib = {},
) -> BaseNode:
  if len(args) < 2:
    return BaseNode(type=NodeType.BOOLEAN, bool_value=False)
  
  result: bool = args[0].fetch_float() >= args[1].fetch_float()
  return BaseNode(
    type=NodeType.BOOLEAN,
    bool_value=result
  )

@register_fn(fn_exports)
@rename_fn("not")
def not_fn(
  args: list[BaseNode],
  _: Node | None = None,
  __: FnLib = {},
) -> BaseNode:
  if len(args) < 1:
    return BaseNode(
      type=NodeType.BOOLEAN,
      bool_value=True
    )
  
  return BaseNode(
    type=NodeType.BOOLEAN,
    bool_value=not args[0].fetch_bool()
  )




from typing import Generator, Callable

build: Callable[
  [Callable[[], Generator[str, None, None]]],
  Generator[str, None, None]
] = lambda fn: fn()

@build
def ghost_id_generator() -> Generator[str, None, None]:
  ghost_id: int = 0

  while True:
    yield f"ghost:{ghost_id}"
    ghost_id += 1

def generate_ghost_id() -> str:
  return next(ghost_id_generator)




from src.models.node_type import NodeType
from src.models.base_node import BaseNode
from src.models.node import Node, FnLib, ExecFn, HOLib, HOExecFn

from src.internals.ghost import generate_ghost_id
from src.internals.registry import (
  register_fn,
  setup as setup_registry
)

fn_exports: list[ExecFn] = []
ho_exports: list[HOExecFn] = []
def setup(fn_lib: FnLib, ho_lib: HOLib) -> None:
  setup_registry(fn_lib, ho_lib, fn_exports, ho_exports)

@register_fn(fn_exports)
def collection(
  args: list[BaseNode],
  _: Node | None = None,
  fn_lib: FnLib = {},
) -> BaseNode:
  ghost_id: str = generate_ghost_id()
  
  def fetch_fn(
    fn_args: list[BaseNode],
    _: Node | None = None,
    __: FnLib = {},
  ) -> BaseNode:
    if len(fn_args) == 0:
      return BaseNode(
        type=NodeType.GHOST,
        ghost_value=ghost_id,
      )
    
    index: int = int(fn_args[0].fetch_float())
    if index < 0 or index >= len(args):
      return BaseNode()
    
    return args[index]
  
  def push_fn(
    fn_args: list[BaseNode],
    _: Node | None = None,
    __: FnLib = {},
  ) -> BaseNode:
    if len(fn_args) == 0:
      fn_args.append(BaseNode())
    
    value: BaseNode = fn_args[0]
    if len(fn_args) == 1:
      args.append(value)
      return BaseNode(
        type=NodeType.BOOLEAN,
        bool_value=True,
      )
    
    index: int = int(fn_args[1].fetch_float())
    if index < 0 or index >= len(args):
      return BaseNode(
        type=NodeType.BOOLEAN,
        bool_value=False,
      )
    
    args.insert(index, value)
    
    return BaseNode(
      type=NodeType.BOOLEAN,
      bool_value=True
    )

  def update_fn(
    fn_args: list[BaseNode],
    _: Node | None = None,
    __: FnLib = {},
  ) -> BaseNode:
    value: BaseNode = fn_args[0]
    index: int = int(fn_args[1].fetch_float())

    if index < 0 or index >= len(args):
      return BaseNode(
        type=NodeType.BOOLEAN,
        bool_value=False,
      )
    
    args[index] = value
    return BaseNode(
      type=NodeType.BOOLEAN,
      bool_value=True
    )

  def pop_fn(
    fn_args: list[BaseNode],
    _: Node | None = None,
    __: FnLib = {},
  ) -> BaseNode:
    fn_args.append(BaseNode(
      type=NodeType.NUMERIC,
      float_value=len(args) - 1
    ))

    index: int = int(fn_args[0].fetch_float())
    if index < 0 or index >= len(args):
      return BaseNode()
    
    result = args.pop(index)
    return result
  
  def size_fn(*_) -> BaseNode:
    return BaseNode(
      type=NodeType.NUMERIC,
      float_value=float(len(args))
    )

  fn_lib[f"{ghost_id}:primary"] = fetch_fn
  fn_lib[f"{ghost_id}:push"] = push_fn
  fn_lib[f"{ghost_id}:pop"] = pop_fn
  fn_lib[f"{ghost_id}:update"] = update_fn
  fn_lib[f"{ghost_id}:size"] = size_fn

  return BaseNode(
    type=NodeType.GHOST,
    ghost_value=ghost_id,
  )

@register_fn(fn_exports)
def push(
  args: list[BaseNode],
  _: Node | None = None,
  fn_lib: FnLib = {},
) -> BaseNode:
  if len(args) == 0:
    return BaseNode(
      type=NodeType.BOOLEAN,
      bool_value=False
    )

  collection_arg = args[0]
  ghost_value: str | None = collection_arg.ghost_value
  if ghost_value is None:
    return BaseNode(
      type=NodeType.BOOLEAN,
      bool_value=False
    )
  
  push_fn: ExecFn | None = fn_lib.get(f"{ghost_value}:push")
  if push_fn is None:
    return BaseNode(
      type=NodeType.BOOLEAN,
      bool_value=False
    )
  
  push_value: BaseNode = BaseNode()
  if len(args) == 2:
    push_value = args[1]

  push_index: int | None = None
  if len(args) > 2:
    push_index = int(args[2].fetch_float())
  
  fn_args: list[BaseNode] = [push_value]
  if push_index is not None:
    fn_args.append(BaseNode(
      type=NodeType.NUMERIC,
      float_value=float(push_index)
    ))
  
  return push_fn(fn_args, None, fn_lib)

@register_fn(fn_exports)
def pop(
  args: list[BaseNode],
  _: Node | None = None,
  fn_lib: FnLib = {},
) -> BaseNode:
  if len(args) == 0: return BaseNode()

  collection_arg = args[0]
  ghost_value: str | None = collection_arg.ghost_value
  if ghost_value is None: return BaseNode()
  
  pop_fn: ExecFn | None = fn_lib.get(f"{ghost_value}:pop")
  if pop_fn is None: return BaseNode()
  
  push_index: int | None = None
  if len(args) > 1:
    push_index = int(args[1].fetch_float())
  
  fn_args: list[BaseNode] = []
  if push_index is not None:
    fn_args.append(BaseNode(
      type=NodeType.NUMERIC,
      float_value=float(push_index)
    ))
  
  return pop_fn(fn_args, None, fn_lib)

@register_fn(fn_exports)
def update(
  args: list[BaseNode],
  _: Node | None = None,
  fn_lib: FnLib = {},
) -> BaseNode:
  if len(args) == 0:
    return BaseNode(
      type=NodeType.BOOLEAN,
      bool_value=False
    )

  collection_arg = args[0]
  ghost_value: str | None = collection_arg.ghost_value
  if ghost_value is None:
    return BaseNode(
      type=NodeType.BOOLEAN,
      bool_value=False
    )
  
  update_fn: ExecFn | None = fn_lib.get(f"{ghost_value}:update")
  if update_fn is None:
    return BaseNode(
      type=NodeType.BOOLEAN,
      bool_value=False
    )
  
  if len(args) < 3:
    return BaseNode(
      type=NodeType.BOOLEAN,
      bool_value=False
    )
  
  update_value = args[1]
  push_index = int(args[2].fetch_float())
  fn_args: list[BaseNode] = [
    update_value, BaseNode(
      type=NodeType.NUMERIC,
      float_value=float(push_index)
    )
  ]
  
  return update_fn(fn_args, None, fn_lib)

@register_fn(fn_exports)
def size(
  args: list[BaseNode],
  _: Node | None = None,
  fn_lib: FnLib = {},
) -> BaseNode:
  if len(args) == 0: return BaseNode()

  collection_arg = args[0]
  ghost_value: str | None = collection_arg.ghost_value
  if ghost_value is None: return BaseNode()
  
  size_fn: ExecFn | None = fn_lib.get(f"{ghost_value}:size")
  if size_fn is None: return BaseNode()
  return size_fn([], None, fn_lib)




from src.models.node import FnLib, HOLib

from .console import setup as setup_clio
from .literal import setup as setup_literals
from .component import setup as setup_component
from .operator import setup as setup_operators
from .loop import setup as setup_loops
from .control import setup as setup_control
from .collection import setup as setup_collection

def setup(fn_lib: FnLib, ho_lib: HOLib) -> None:
  setup_clio(fn_lib, ho_lib)
  setup_literals(fn_lib, ho_lib)
  setup_component(fn_lib, ho_lib)
  setup_operators(fn_lib, ho_lib)
  setup_loops(fn_lib, ho_lib)
  setup_control(fn_lib, ho_lib)
  setup_collection(fn_lib, ho_lib)




from src.models.node_type import NodeType
from src.models.base_node import BaseNode
from src.models.node import Node, FnLib, ExecFn, HOLib, HOExecFn

from src.internals.registry import (
  register_fn,
  setup as setup_registry
)

fn_exports: list[ExecFn] = []
ho_exports: list[HOExecFn] = []
def setup(fn_lib: FnLib, ho_lib: HOLib) -> None:
  setup_registry(fn_lib, ho_lib, fn_exports, ho_exports)

@register_fn(fn_exports)
def base(
  _: list[BaseNode],
  node: Node | None = None,
  __: FnLib = {},
) -> BaseNode:
  if (node is None): return BaseNode()
  if (node.value is None): return BaseNode()
  
  return node.value

@register_fn(fn_exports)
def endline(*_) -> BaseNode:
  return BaseNode(
    type=NodeType.TEXT,
    str_value="\n"
  )

@register_fn(fn_exports)
def numeric(
  args: list[BaseNode],
  *_,
) -> BaseNode:
  args.append(BaseNode(type=NodeType.NUMERIC, float_value=0.0))
  return BaseNode(
    type=NodeType.NUMERIC,
    float_value=args[0].fetch_float()
  )

@register_fn(fn_exports)
def boolean(
  args: list[BaseNode],
  *_,
) -> BaseNode:
  args.append(BaseNode(type=NodeType.BOOLEAN, float_value=False))
  return BaseNode(
    type=NodeType.BOOLEAN,
    bool_value=args[0].fetch_bool()
  )

@register_fn(fn_exports)
def text(
  args: list[BaseNode],
  *_,
) -> BaseNode:
  args.append(BaseNode(type=NodeType.TEXT, str_value=""))
  return BaseNode(
    type=NodeType.TEXT,
    str_value=args[0].fetch_str()
  )




from __future__ import annotations
from typing import Self, TypeAlias, Callable

from .node_type import NodeType
from .base_node import BaseNode
from .node_signal import NodeSignal

class Node(BaseNode):
  executes: str | None
  script: Node | None
  value: BaseNode | None = None
  next: Node | None = None
  high_order: bool = False

  def __init__(
    self: Self,
    executes: str | None = None,
    script: Node | None = None,
    value: BaseNode | None = None,
    next_node: Node | None = None,
    high_order: bool = False,
  ) -> None:
    super().__init__(
      type=NodeType.NODE,
      base=False,
    )

    self.executes = executes
    self.script = script
    self.value = value
    self.next = next_node
    self.high_order = high_order

  def high_order_execute(
    self: Self,
    fn_lib: FnLib,
    ho_lib: HOLib,
  ) -> Node | None:
    if self.executes is None: return None
    if self.executes not in ho_lib: return None

    return ho_lib[self.executes](
      self,
      fn_lib,
      ho_lib,
    )

  def execute(
    self: Self,
    fn_lib: FnLib,
    ho_lib: HOLib,
  ) -> tuple[BaseNode, Node | None]:
    if self.high_order:
      next_node = self.high_order_execute(fn_lib, ho_lib)
      if next_node is None: return BaseNode(), self.next
      return next_node.execute(
        fn_lib,
        ho_lib
      )

    current: Node | None = self.script
    args: list[BaseNode] = []
    result: BaseNode = BaseNode()

    while current is not None:
      value, next_node = current.execute(fn_lib, ho_lib)
      args.append(value)

      if value.signal == NodeSignal.RETURN:
        result = value
        next_node = None

      current = next_node

    if self.executes is None:
      return result, self.next
    
    result = fn_lib[self.executes](
      args,
      self,
      fn_lib
    )

    return result, self.next

ExecFn: TypeAlias = Callable[
  [
    list[BaseNode],
    Node | None,
    dict[str, "ExecFn"]
  ],
  BaseNode
]

HOExecFn: TypeAlias = Callable[
  [
    Node,
    dict[str, ExecFn],
    dict[str, "HOExecFn"]
  ],
  Node | None
]

FnLib: TypeAlias = dict[str, ExecFn]
HOLib: TypeAlias = dict[str, HOExecFn]



from src.models.node_type import NodeType
from src.models.node_signal import NodeSignal
from src.models.base_node import BaseNode
from src.models.node import Node, FnLib, ExecFn, HOLib, HOExecFn

from src.internals.registry import (
  register_fn,
  register_ho,
  setup as setup_registry
)

from .collection import collection

fn_exports: list[ExecFn] = []
ho_exports: list[HOExecFn] = []
def setup(fn_lib: FnLib, ho_lib: HOLib) -> None:
  setup_registry(fn_lib, ho_lib, fn_exports, ho_exports)

@register_fn(fn_exports)
def script(
  args: list[BaseNode],
  _: Node | None = None,
  __: FnLib = {},
) -> BaseNode:
  for arg in args:
    if arg.signal == NodeSignal.RETURN:
      return arg
  return BaseNode()

@register_ho(ho_exports)
def component(
  node: Node,
  fn_lib: FnLib,
  ho_lib: HOLib,
) -> Node | None:
  name_node = node.script
  if name_node is None: return None
  
  component_name, args_node = name_node.execute(fn_lib, ho_lib)
  if component_name.type != NodeType.TEXT:
    return None
  
  if args_node is None:
    return None
  
  args_name, script_node = args_node.execute(fn_lib, ho_lib)
  if args_name.type != NodeType.TEXT:
    return None
  
  def component_fn(
    args: list[BaseNode],
    _: Node | None = None,
    fn_lib: FnLib = {},
  ) -> BaseNode:
    if not script_node:
      return BaseNode()
    
    local_fn_scope: FnLib = {}
    for fn_keys in fn_lib:
      local_fn_scope[fn_keys] = fn_lib[fn_keys]

    saves([
      BaseNode(
        type=NodeType.TEXT,
        str_value=args_name.fetch_str()
      ),
      collection(args, None, local_fn_scope),
    ], None, local_fn_scope)
    
    result: BaseNode = script_node.execute(
      local_fn_scope,
      ho_lib,
    )[0]

    result.signal = None
    return result
  
  fn_lib[component_name.fetch_str()] = component_fn
  
  return node.next

@register_fn(fn_exports)
def saves(
  args: list[BaseNode],
  _: Node | None = None,
  fn_lib: FnLib = {},
) -> BaseNode:
  if len(args) <= 1:
    return BaseNode()

  name: str = args[0].fetch_str()
  value: BaseNode = args[1]

  def variable_fn(
    fn_args: list[BaseNode],
    node: Node | None = None,
    fn_lib: FnLib = {},
  ) -> BaseNode:
    if value.type != NodeType.GHOST:
      return value
    
    ghost_value = value.ghost_value
    if f"{ghost_value}:primary" not in fn_lib:
      return value
    
    return fn_lib[f"{ghost_value}:primary"](
      fn_args,
      node,
      fn_lib
    )

  fn_lib[name] = variable_fn
  return value

@register_fn(fn_exports)
def returns(
  args: list[BaseNode],
  _: Node | None = None,
  __: FnLib = {},
) -> BaseNode:
  if len(args) == 0:
    return BaseNode()
  
  args[0].signal = NodeSignal.RETURN
  return args[0]




from __future__ import annotations

from .models.node import FnLib, HOLib, Node
from .components import setup
from .internals.parser import parse

fn_lib: FnLib = {}
ho_lib: HOLib = {}

setup(fn_lib, ho_lib)

def run_source(source: str) -> None:
  source_lines: list[str] = source.split("\n")
  root_script: Node = parse(source_lines, set(ho_lib.keys()))
  root_script.execute(fn_lib, ho_lib)




from src.models.node_type import NodeType
from src.models.node import ExecFn, HOExecFn, FnLib, HOLib
from src.models.base_node import BaseNode

from src.internals.registry import (
  register_fn,
  rename_fn,
  setup as setup_registry
)

fn_exports: list[ExecFn] = []
ho_exports: list[HOExecFn] = []
def setup(fn_lib: FnLib, ho_lib: HOLib) -> None:
  setup_registry(fn_lib, ho_lib, fn_exports, ho_exports)

@register_fn(fn_exports)
def output(
  args: list[BaseNode],
  *_,
) -> BaseNode:
  if not args:
    return BaseNode()
  
  converted_args: list[str] = [
    arg.fetch_str()
    for arg in args
  ]

  print(
    *converted_args,
    sep="",
    end=""
  )

  return BaseNode()

@register_fn(fn_exports)
@rename_fn("input")
def input_fn(
  args: list[BaseNode],
  *_,
) -> BaseNode:
  if not args:
    return BaseNode()
  
  prompt: str = ""
  if len(args) >= 1:
    prompt = args[0].fetch_str()

  input_value = input(prompt)

  return BaseNode(
    type=NodeType.TEXT,
    str_value=input_value
  )




from typing import Self

from .node_type import NodeType

class BaseNode:
  type: str
  base: bool
  str_value: str | None
  float_value: float | None
  bool_value: bool | None
  ghost_value: str | None = None
  signal: str | None = None

  def __init__(
    self: Self,
    base: bool = True,
    type: str = NodeType.VOID,
    str_value: str | None = None,
    float_value: float | None = None,
    bool_value: bool | None = None,
    ghost_value: str | None = None,
    signal: str | None = None,
  ) -> None:
    self.base = base
    self.type = type
    self.signal = signal

    self.str_value = str_value
    self.float_value = float_value
    self.bool_value = bool_value

    self.ghost_value = ghost_value

  def fetch_str(self: Self) -> str:
    if self.type == NodeType.NUMERIC:
      return str(self.fetch_float())
    
    if self.type == NodeType.BOOLEAN:
      return 'true' if self.fetch_bool() else 'false'
    
    if self.str_value is None:
      return ""
    
    return self.str_value

  def fetch_float(self: Self) -> float:
    if self.type == NodeType.BOOLEAN:
      return 1.0 if self.fetch_bool() else 0.0
    
    if self.type == NodeType.TEXT:
      try:
        return float(self.fetch_str())
      except ValueError:
        return 0.0
      
    if self.float_value is None:
      return 0.0
    
    return self.float_value
  
  def fetch_bool(self: Self) -> bool:
    if self.type == NodeType.NUMERIC:
      return self.fetch_float() != 0.0
    
    if self.type == NodeType.TEXT:
      return len(self.fetch_str()) != 0
    
    if self.bool_value is None:
      return False
    
    return self.bool_value




from typing import Self
from sys import argv

from src.core import run_source

class CLIOption:
  option: str
  meta: str | None = None

  def __init__(
    self: Self,
    option: str = "",
    meta: str | None = None,
  ) -> None:
    self.option = option
    self.meta = meta
  
def run_file(file: str) -> None:
  source: str = ""

  try:
    with open(file, "r") as source_file:
      source = source_file.read()
      source_file.close()
  except FileNotFoundError:
    print(f"Error: File '{file}' not found.")
  
  run_source(source)

def run_options(options: list[CLIOption]) -> None:
  run_file(options[0].option)

def main() -> None:
  cli_options: list[CLIOption] = []
  meta: str | None = None

  for arg in argv[1:]:
    if arg.startswith("-"):
      meta = arg[1:]
      continue

    cli_options.append(CLIOption(
      option=arg,
      meta=meta
    ))
    meta = None
  else:
    if meta is not None:
      cli_options.append(CLIOption(
        option="",
        meta=meta
      ))

  run_options(cli_options)




from src.models.node import Node, FnLib, ExecFn, HOLib, HOExecFn

from src.internals.registry import (
  register_ho,
  setup as setup_registry
)

fn_exports: list[ExecFn] = []
ho_exports: list[HOExecFn] = []
def setup(fn_lib: FnLib, ho_lib: HOLib) -> None:
  setup_registry(fn_lib, ho_lib, fn_exports, ho_exports)

@register_ho(ho_exports)
def loop(
  node: Node,
  fn_lib: FnLib,
  ho_lib: HOLib,
) -> Node | None:
  condition_node = node.script
  if condition_node is None: return node.next

  condition, loop_node = condition_node.execute(fn_lib, ho_lib)
  
  while condition.fetch_bool() and loop_node:
    loop_node.execute(fn_lib, ho_lib)
    condition, loop_node = condition_node.execute(fn_lib, ho_lib)
  
  return node.next




from src.models.base_node import BaseNode
from src.models.node_type import NodeType
from src.models.node import Node

def identify(content: str) -> str:
  if len(content) > 2 and content[0] == '"' and content[-1] == '"':
    return NodeType.TEXT
  
  if content[0] == ":":
    return NodeType.TEXT
  
  nums: int = 0
  decimals: int = 0
  hyphens: int = 0

  for char in content:
    if char.isdigit():
      nums += 1
    elif char == '.':
      decimals += 1
    elif char == '-':
      hyphens += 1
    else:
      break
  else:
    if decimals > 1: return NodeType.NODE
    if hyphens > 1: return NodeType.NODE
    if hyphens == 1 and content[0] != '-': return NodeType.NODE
    if nums == 0: return NodeType.NODE
    
    return NodeType.NUMERIC

  if content.lower() in ['true', 'false']:
    return NodeType.BOOLEAN
  
  return NodeType.NODE

def convert(
  content: str,
  high_order_list: set[str]
) -> Node:
  content_type: str = identify(content)
  if content_type == NodeType.NODE:
    return Node(
      executes=content,
      high_order=content in high_order_list,
    )
  
  str_value: str | None = None
  float_value: float | None = None
  bool_value: bool | None = None

  if content_type == NodeType.TEXT:
    if content[0] == ":":
      str_value = content[1:]
    else:
      str_value = content[1:-1]
  if content_type == NodeType.NUMERIC:
    float_value = float(content)
  
  if content_type == NodeType.BOOLEAN:
    bool_value = content.lower() == 'true'
  
  return Node(
    executes="base",
    value=BaseNode(
      type=content_type,
      str_value=str_value,
      float_value=float_value,
      bool_value=bool_value,
    )
  )

def parse(
  lines: list[str],
  high_order_list: set[str]
) -> Node:
  levels: list[Node] = [Node(executes="script")]
  level = [-1]

  def dive(node: Node) -> None:
    levels[-1].script = node
    levels.append(node)
    level[0] += 1

  def overtake(node: Node) -> None:
    levels[-1].next = node
    levels.pop()
    levels.append(node)

  def fly(node: Node, node_level: int) -> None:
    while node_level < level[0]:
      levels.pop()
      level[0] -= 1

    overtake(node)

  for line in lines:
    current_level: int = 0
    
    i = 0
    while i < len(line) and line[i] == ' ':
      current_level += 1
      i += 1
    
    content: str = line[i:]
    if not content:
      continue

    node: Node = convert(content, high_order_list)
    
    if current_level == level[0]:
      overtake(node)
    elif current_level > level[0]:
      dive(node)
    else:
      fly(node, current_level)

  return levels[0]




from setuptools import setup, find_packages
from pathlib import Path

def read_config(file: str) -> str:
  return Path(file).read_text().strip()

setup(
  name='gsam',
  include_package_data=True,
  version=read_config("VERSION"),
  packages=find_packages(),
  author='AttAditya',
  description='A compiler for the GSAM language.',
  long_description=read_config("README.md"),
  long_description_content_type='text/markdown',
  url='https://github.com/AttAditya/gsam-compiler',
  python_requires='>=3.8',
  entry_points={
    'console_scripts': [
      'gsamc=src.cli:main',
      'gsam=src.cli:main',
      'gsam-compiler=src.cli:main',
      'gsam-compiler-cli=src.cli:main',
      'gsamc-cli=src.cli:main',
    ],
  }
)




from src.models.node import HOExecFn, ExecFn, FnLib, HOLib, Node

from src.internals.registry import (
  register_ho,
  rename_ho,
  setup as setup_registry
)

fn_exports: list[ExecFn] = []
ho_exports: list[HOExecFn] = []
def setup(fn_lib: FnLib, ho_lib: HOLib) -> None:
  setup_registry(fn_lib, ho_lib, fn_exports, ho_exports)

@register_ho(ho_exports)
@rename_ho("if")
def if_fn(
  node: Node,
  fn_lib: FnLib,
  ho_lib: HOLib,
) -> Node | None:
  condition_node = node.script
  if condition_node is None: return node.next

  condition, true_node = condition_node.execute(fn_lib, ho_lib)
  if true_node is None: return node.next
  false_node: Node | None = true_node.next

  true_node = Node(
    executes= true_node.executes,
    script= true_node.script,
    next_node=node.next,
  )
  if false_node is not None:
    false_node = Node(
      executes= false_node.executes,
      script= false_node.script,
      next_node=node.next,
    )
  
  if condition.fetch_bool():
    return true_node
  
  return false_node

